/*
 *   Copyright (c) 2010, Michael Lehn
 *
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1) Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2) Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *   3) Neither the name of the FLENS development group nor the names of
 *      its contributors may be used to endorse or promote products derived
 *      from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <flens/storage/storageinfo.h>

namespace flens { namespace blas {

#ifdef HAVE_CXXBLAS_HERK

//-- herk
template <typename ALPHA, typename MA, typename BETA, typename MC>
void
rk(cxxblas::Transpose trans,
   const ALPHA &alpha,
   const GeMatrix<MA> &A,
   const BETA &beta,
   HeMatrix<MC> &C)
{
    ASSERT(StorageInfo<MC>::Order==StorageInfo<MA>::Order);
    typedef typename HeMatrix<MC>::IndexType IndexType;

    IndexType n = (trans==cxxblas::NoTrans) ? A.numRows()
                                            : A.numCols();
    IndexType k = (trans==cxxblas::NoTrans) ? A.numCols()
                                            : A.numRows();

    ASSERT((beta==static_cast<BETA>(0)) || (C.dim()==n));
    if (C.dim()!=n) {
        C.engine().resize(n, n);
    }

    ASSERT(C.dim()==((trans==cxxblas::NoTrans) ? A.numRows() : A.numCols()));
    cxxblas::herk(StorageInfo<MC>::Order, C.upLo(),
                  trans, n, k, alpha,
                  A.engine().data(), A.engine().leadingDimension(),
                  beta,
                  C.engine().data(), C.engine().leadingDimension());
}

#endif // HAVE_CXXBLAS_HERK

#ifdef HAVE_CXXBLAS_SYRK

//-- syrk
template <typename ALPHA, typename MA, typename BETA, typename MC>
void
rk(cxxblas::Transpose trans,
   const ALPHA &alpha,
   const GeMatrix<MA> &A,
   const BETA &beta,
   SyMatrix<MC> &C)
{
    ASSERT(StorageInfo<MC>::Order==StorageInfo<MA>::Order);

    typedef typename SyMatrix<MC>::IndexType IndexType;

    IndexType n = (trans==cxxblas::NoTrans) ? A.numRows()
                                            : A.numCols();
    IndexType k = (trans==cxxblas::NoTrans) ? A.numCols()
                                            : A.numRows();

    ASSERT((beta==static_cast<BETA>(0)) || (C.dim()==n));
    if (C.dim()!=n) {
        C.engine().resize(n, n);
    }

    ASSERT(C.dim()==((trans==cxxblas::NoTrans) ? A.numRows() : A.numCols()));
    cxxblas::syrk(StorageInfo<MC>::Order, C.upLo(),
                  trans, n, k, alpha,
                  A.engine().data(), A.engine().leadingDimension(),
                  beta,
                  C.engine().data(), C.engine().leadingDimension());
}

#endif  // HAVE_CXXBLAS_SYRK

} } // namespace blas, flens
